#ifndef P1CAMERA_TYPES_HPP
#define P1CAMERA_TYPES_HPP

#include <cstdint>
#include <string>
#include <vector>
#include <functional>

#include "C_P1CameraCommonTypes.h"
#include "C_P1CameraCommonLinkage.h"

namespace P1
{
namespace CameraSdk
{

/**
 * @brief Holds a single property value, and acts like a wrapper of different property types
 *
 * This class holds only the value of a property. It does not know about the property itself,
 * for that we have the @ref PropertySpecification class. PropertyValue can hold value of type:
 * `integer`, `floating point`, `string`, `point`, `color`, `area` and `rect`.
 *
 * A propertyValue also have a @ref PropertyValue::mPresentationString, that is a string
 * representation of its value.
 *
 * Preferrably you should not create PropertyValues yourself. Rather you should obtains them
 * from a camera and modify them.
 *
 * If you are familiar with *Qt*, the `QVariant` class achieves the same goal as this class.
 *
 * @sa PropertyType
 * @sa Camera::GetProperty
 */
struct EXPORT_API PropertyValue
{
    /** @brief Create an empty property value (no data) */
    PropertyValue() { }

    /** @brief Create a string type property value, based on the provided string */
    PropertyValue(std::string const& fromString);

    /** @brief Create a string type property value, based on the provided C string */
    PropertyValue(char const* fromCString);

    /** @brief Create a Int64 type property value, based on the provided integer */
    PropertyValue(int64_t const& fromInteger);

    /** @brief Create a Int32 type property value, based on the provided integer */
    PropertyValue(int32_t const& fromInteger);

    /** @brief Create a UInt32 type property value, based on the provided unsigned integer */
    PropertyValue(uint32_t const& fromUInteger);

    /** @brief Create a UInt64 type property value, based on the provided unsigned integer */
    PropertyValue(uint64_t const& fromUInteger);

    /** @brief Create a float64 type property value, based on the provided double */
    PropertyValue(double const& fromDouble);

    /** @brief Indicates the type of the value contained in this PropertyValue object */
    PropertyType mType = kPropertyTypeUndefined;

    /** @brief All integer based values are contained in this member (enum, int, uint, time, etc.) */
    int64_t mInt = 0;

    /** @brief All floating point values are contained in this member. (float 32 and 64) */
    double mDouble = 0.0;

    /** @brief If this PropertyValue is a string type, the string is contained here */
    MSVC_ALLOW_PUBLIC_STL(std::string mString);

    /** @brief If this PropertyValue is a point type, it is contained here */
    ValuePoint mPoint;

    /** @brief If this PropertyValue is a color type, it is contained here */
    ValueColorFloat mColor;

    /** @brief If this PropertyValue is a area type, it is contained here */
    ValueArea mArea;

    /** @brief If this PropertyValue is a floating point area type, it is contained here */
    ValueAreaFloat mAreaFloat;

    /** @brief If this PropertyValue is a rectangle type, it is contained here */
    ValueRect mRect;

    /** @brief If this PropertyValue is a integer range type, it is contained here */
    ValueRange mRange { 0, 0 };

    /** @brief If this PropertyValue is a floating point range type, it is contained here */
    ValueRangeFloat mRangeFloat { 0.0, 0.0 };

    /** @brief This is a human readable representation of the value */
    MSVC_ALLOW_PUBLIC_STL(std::string mPresentationString);

    /** @brief Return the this property's value as an unsigned Int32 */
    uint32_t IntegerAsU32() const;

    /** @brief Return the this property's value as a signed Int32 */
    int32_t IntegerAsI32() const;

    /** @brief Return the this property's value as an unsigned Int64 */
    uint64_t IntegerAsU64() const;

    /** @brief Return the this property's value as a signed Int64 */
    int64_t IntegerAsI64() const;

    /**
     * @brief Returns a simple string representation of this property
     *
     * This method uses the "RepresentationString", if present.
     *
     * @return A string value that describes the PropertyValue's value
     */
    std::string StringValue() const;

    /**
     * @brief Returns a description of the property value, including its type name
     *
     * This method is equivalent to using the std::ostream << operator
     */
    std::string ToString() const;

    /**
     * @brief Return a human readable name of this property's type
     */
    std::string TypeString() const;

    /**
     * @brief Create a enum type property value, based on the provided integer
     */
    static PropertyValue CreateEnum(uint32_t enumValue);

    /** @brief Compare type and value */
    bool operator==(PropertyValue const& other) const;
    bool operator!=(PropertyValue const& other) const;
};

EXPORT_API std::ostream& operator<<(std::ostream& os, PropertyValue const& val);

// C++ Version of PropertyRange

/**
 * @brief Numeric property _range span_ class, used by PropertySpecification
 *
 * Some camera properties have a valid range of values, defined by a minimum and
 * maximum. This class defines these boundaries.
 *
 * A valid PropertyRange object, _must_ have its member @ref mHas set to `true`.
 * `false` indicates an invalid range object.
 */
struct EXPORT_API PropertyRange
{
    /** @brief Indicate if this object is valid, and range check should be applied */
	bool mHas = false;

    /** @brief The lower bound of the range */
	PropertyValue mMin;

    /** @brief The upper bound of the range */
	PropertyValue mMax;

    /** @brief Get a human readable representation of the range */
	std::string ToString() const;
};

std::ostream& operator<<(std::ostream& os, PropertyRange const& range);


/**
 * @brief A detailed description of a camera property
 *
 * This class defines if a property has any restrictions to its valid values,
 * like a range or even a list of allowed values.
 *
 * All Properties belong to a category or group, have a unit and a visibility
 * level, called the _user level_. The _user level_ defines if the property
 * is an advanced camera setting, that should probably be hidden from novice
 * users.
 *
 * To represent the current value of a proerty we use the class
 * @ref PropertyValue. It is also there the _type_ of the property is defined.
 *
 * You cannot change property specifications, you can change only property
 * values, and only using @ref Camera::SetProperty
 *
 * You can obtain PropertySpecification objects from @ref Camera::PropertySpec
 */
struct EXPORT_API PropertySpecification
{
    /**
     * @brief The assigned id of the property
     *
     * All properties are identified uniquely by their _Id_. You use the _Id_ to
     * reference and look up properties on a camera.
     */
	uint32_t mId = kPropertyUndefined;

    /**
     * @brief The current value (and type) of the property
     *
     * This is the current assigned value to the property. If the property if
     * not marked as _read-only_ (@ref mReadOnly), this value can be changed using
     * @ref Camera::SetProperty.
     */
	PropertyValue mValue;

    /**
     * @brief The (human readable) name of the property
     *
     * This name is localized based on the camera's UI language setting. However,
     * the default language is english.
     *
     * This name can be presented to users in a UI
     */
    MSVC_ALLOW_PUBLIC_STL(std::string) mName;

    /**
     * @brief The group that this property belongs to
     *
     * The group is a string that might contain double colons "::", to denote
     * subgroups. You can use the _group_ to group related properties in a UI.
     */
    MSVC_ALLOW_PUBLIC_STL(std::string) mGroup;

    /**
     * @brief Indicates if this property is currently _read only_
     *
     * Some properties are always _read only_, but others might be _read only_
     * as a result of the camera settings. For example, if the exposure program
     * is set to _auto_, then the _ISO_, _Aperture_ and _shutter_ properties are
     * _read_only_. If the exposure program is set to _manual_, then they are
     * writeable.
     */
	bool mReadOnly = false;

    /**
     * @brief
     *
     *
     */
	bool mEnabled = true;

    /**
     * @brief The visibility level of this peroperty
     *
     * Some properties are more advanced than others. Use this member to hide
     * the advanced properties from novice users.
     */
	PropertyUserLevel mUserLevel = kPropertyUserLevelUndefined;

    /**
     * @brief
     *
     *
     */
	bool mSelectOnly = false;

    /**
     * @brief
     *
     *
     */
	bool mDisplayAsHex = false;

    /**
     * @brief
     *
     *
     */
	bool mRefreshRequired = false;

    /**
     * @brief
     *
     *
     */
    MSVC_ALLOW_PUBLIC_STL(std::string mUnitString);

    /**
     * @brief
     *
     *
     */
    MSVC_ALLOW_PUBLIC_STL(std::string mDescription);

    /**
     * @brief The sorting order of the property
     *
     * Instead of presenting properties as sorted by _Id_ or name, sort by
     * this integer. (Per group.)
     */
	uint32_t mDisplayIndex = 0;

    /**
     * @brief The default value of the property
     *
     * This is the value that the property is restored to, then "reset to
     * default" is triggered, on the camera.
     *
     * _Note_: that not all properties have a meaningful default value.
     * (Such as "System clock")
     */
	PropertyValue mDefault;

    /**
     * @brief The valid range of the property value
     *
     * Some properties are restricted to have a valid range of values they can
     * take.
     * If a range is applicable for this property, the @ref PropertyRange::mHas
     * is `true`. If `false`, this member can be ignored.
     */
	PropertyRange mRange;

    /**
     * @brief List of allowed values, for this property
     *
     * Some properties can only take values, from this list of allowed values.
     * If this list is present (size > 0), and you set a value that is not in
     * the list, two things can happen: 1. the camera rejects the value or, 2.
     * the value is truncated to the nearest value in the list.
     */
    MSVC_ALLOW_PUBLIC_STL(std::vector<PropertyValue> mValueList);

    /**
     * @brief Get a human readable description of this PropertySpecification
     */
	std::string ToString() const;
};

std::ostream& operator<<(std::ostream& os, PropertySpecification const& spec);

/**
 * @brief Brief description of a camera - without necessarily having to open it
 *
 * This class is used by the @ref Camera::GetAvailableCameras method, to
 * return a list of available cameras. You do not have to create instances of this
 * class yourself.
 *
 * @sa Camera::GetAvailableCameras
 */
struct EXPORT_API CameraListElement
{
    /** The connection type where camera is present */
    TetherConnection mConnection = kTetherConnection_Undefined;

    /** Camera display name */
    MSVC_ALLOW_PUBLIC_STL(std::string mName);

    /** The cameras serial number */
    MSVC_ALLOW_PUBLIC_STL(std::string mSerialNum);

    /** Camera's IP address, if it is available via TCP/IP */
    MSVC_ALLOW_PUBLIC_STL(std::string mIPv4);

    /**
     * @brief Bool indicating if the camera is open by the SDK.
     *
     * Note, this is a snapshot of the state when this camera-list-element
     * is created, hence may have changed since.
     */
    bool mIsOpen = false;
};

std::ostream& operator<<(std::ostream& os, CameraListElement const& elem);

//class Camera; // forward decl

} // namespace CameraSdk
} // namespace P1

#endif // P1CAMERA_TYPES_HPP
