
#ifndef C_P1CameraCommonEventTypes_h
#define C_P1CameraCommonEventTypes_h

#include <string>
#include <stdint.h>
#include "C_P1CameraCommonLinkage.h"

namespace P1
{
namespace CameraSdk
{
    /**
     * @brief Base class for Notification EventTypes. All event categories inherit from this class
     *
     * The BaseEventType is implemented as an *Enumeration Class*, as described here:
     * https://t.ly/KUa3 (or Google "microsoft enumeration classes")
     *
     * This concept is that instead of using a tranditional enum, we can get better compile time type cheking
     * of the enum values, by implementing them as subclasses of a base class. (Each enum value is a
     * specific subclass.)
     *
     * This allows us to do better compile time checks of which enum values are provided to functions as
     * arguments. As an additional feature each value has human readable string representation.
     */
    class EXPORT_API BaseEventType
    {
    public:
        enum class Category
        {
            Invalid,
            Camera,
            Sdk,
            Property,
            Transfer,
            Image 
        };

        // id is needed for C# bindings
        uint32_t const id;
        MSVC_ALLOW_PUBLIC_STL(std::string const name);
        Category const category;

        bool operator==(BaseEventType const& other) const;
        bool operator!=(BaseEventType const& other) const;
        
        /** Dummy method to make class polymorphic (needed for downcast derived classes to this base class) */
        virtual void dummy();
        virtual ~BaseEventType() = default;

    protected:
        BaseEventType(uint32_t const id,
                      std::string const name,
                      Category const category);
    
    private:
        BaseEventType& operator=(BaseEventType const& other) = delete;
        BaseEventType(BaseEventType const& other) = delete;
    };

    EXPORT_API std::ostream& operator<<(std::ostream& os, BaseEventType const& baseEvent);

// ----------------------------------
// Standard Events
// ----------------------------------

    /**
     * @brief All EventTypes that are not Property type events
     */
    class EXPORT_API EventType : public BaseEventType
    {
    public:
        static EventType const Invalid;                        /**< An invalid event  */
        static EventType const ImageTransferBegin;             /**< The camera has started tramsfering an image (0% complete) */
        static EventType const ImageTransferProgress;          /**< A progress update on the current image transfer (+1% or more) */
        static EventType const CameraImageReady;               /**< An image data ready to be read, as result of previous capture */
        static EventType const CameraDisconnected;             /**< Camera disconnected and therefore unavailable */
        static EventType const CameraLiveViewStarted;          /**< LiveView system will begin to feeding images */
        static EventType const CameraLiveViewStopped;          /**< LiveView system stopped feeding images, and is powered down. */
        static EventType const CameraNewLiveViewImage;         /**< A new LiveView image has arrived and is ready to be consumed by you */
        static EventType const CameraPropertyAdded;            /**< A new property became present on the camera */
        static EventType const CameraPropertyRemoved;          /**< A property was removed from the camera */
        static EventType const CameraImageAdded;               /**< An image was added to the image list */
        static EventType const CameraImageRemoved;             /**< An image was removed to the image list  */
        static EventType const CameraImageEdited;              /**< An image was edited in the image list (changed rating, name, etc.) */
        static EventType const CameraImageNewList;             /**< A new image list was received */
        static EventType const CameraPreviewReady;             /**< An image buffer is ready (type defined by IImageBuffer::Type, e.g., Preview, Tile) */
        static EventType const CameraTileReady;                /**< An image buffer is ready (type defined by IImageBuffer::Type, e.g., Preview, Tile) */
        static EventType const CameraImageBasicImageInfo;      /**< A response to a basic image meta data request */
   
        static EventType const FirmwareTransferBegin;          /**< Transfer of firmware package to camera has begun (0% complete) */
        static EventType const FirmwareTransferProgress;       /**< Transfer of firmware package is now 'x'% complete */
        static EventType const FirmwareTransferDone;           /**< Transfer of firmware package to camera is now 100% completed */
        static EventType const FirmwareTransferError;          /**< Transfer of firmware package to camera has encountered an error. Re-check connection and firmware file location before trying again. */
        
    
    private:
        EventType(uint32_t const id,
                  std::string const name,
                  Category const category);

        EventType(EventType const& other);
        EventType& operator=(EventType const& other);
    };

// ----------------------------------
// Property Events
// ----------------------------------

    /** @brief EventTypes that relate to properties changing */
    class EXPORT_API PropertyEventType : public BaseEventType
    {
    public:
        static PropertyEventType const PropertyValueChange;         /**< Notification describes a property value change event */
        static PropertyEventType const PropertySpecificationChange; /**< Notification describes that a PropertySpecification has changed (not its value) */
    
    private:
        PropertyEventType(uint32_t const id, std::string const name);
        PropertyEventType(PropertyEventType const& other) = delete;
        PropertyEventType& operator=(PropertyEventType const& other) = delete;

    };

// ----------------------------------
// SDK Events
// ----------------------------------

    /** @brief EventTypes that relate to the SDK, independent of a specific camera */
    class EXPORT_API SdkEventType : public BaseEventType
    {
    public:
        /** The thread blocked in WaitForNotification was woken up another threads call to WakeUpWaitingThread */
        static SdkEventType const SdkWakeUp;
        static SdkEventType const SdkCameraFound;    /**< A camera was connected to the computer */
        static SdkEventType const SdkCameraRemoved;  /**< A camera was removed from the computer */
        static SdkEventType const SdkCameraOpened;   /**< A camera was opened by the SDK (on this computer) */
        static SdkEventType const SdkCameraClosed;   /**< A camera was closed by the SDK (on this computer) */

    private:
        SdkEventType(uint32_t const id, std::string const name);
        SdkEventType(SdkEventType const& other);
        SdkEventType& operator=(SdkEventType const& other);
    };
}
}

#endif /* C_P1CameraCommonEventTypes_h */
